#!/usr/bin/env python

"""
High-level classes.

Copyright (C) 2009, 2010, 2011 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from iixr.filesystem import *
from itermerge import itermerge
from os import mkdir    # index discovery
from os.path import exists
import operator

# Constants.

FLUSH_INTERVAL    = 10000
OPEN_PARTITIONS   = 20

# High-level classes.

class Document:

    "A container of document information."

    def __init__(self, docnum):
        self.docnum = docnum
        self.terms = {}

    def add_position(self, term, position):

        """
        Add a position entry for the given 'term', indicating the given
        'position'.
        """

        self.terms.setdefault(term, []).append(position)

class IndexWriter:

    "Building term information and writing it to the term dictionary."

    def __init__(self, pathname, flush_interval):
        self.pathname = pathname
        self.flush_interval = flush_interval

        self.term_partition = get_next_partition(get_term_partitions(self.pathname))

        self.terms = {}
        self.doc_counter = 0

    def add_document(self, doc):

        """
        Add the given document 'doc', updating the document counter and flushing
        terms and fields if appropriate.
        """

        docnum = doc.docnum

        for term, positions in doc.terms.items():
            self.terms.setdefault(term, {})[docnum] = positions

        self.doc_counter += 1

        if self.flush_interval and self.doc_counter >= self.flush_interval:
            self.flush_terms()
            self.doc_counter = 0

    def get_term_writer(self):

        "Return a term writer for the current partition."

        return get_term_writer(self.pathname, self.term_partition)

    def flush_terms(self):

        "Flush terms into the current term partition."

        # Get the terms in order.

        term_writer = self.get_term_writer()
        try:
            term_writer.write_terms(self.terms)
        finally:
            term_writer.close()

        self.terms = {}
        self.term_partition += 1

    def close(self):
        if self.terms or not get_term_partitions(self.pathname):
            self.flush_terms()

class IndexReader(itermerge):

    "Accessing the term dictionaries."

    def __init__(self, pathname, get_reader=None, combine=None):

        # Get the partitions in order.

        partitions = list(get_term_partitions(pathname))
        partitions.sort()

        # Initialise the underlying term partition readers.

        self.readers = [(get_reader or get_term_reader)(pathname, partition) for partition in partitions]
        self.combine = combine or operator.add

        # Initialise this object as an iterator over the readers.

        itermerge.__init__(self, self.readers)
        self.next_value = None

    def get_sizes(self):

        # Readers must have compatible sizes.

        if self.readers:
            return self.readers[0].get_sizes()
        else:
            return 0, 0

    def next(self):
        if self.next_value is not None:
            term, positions = self.next_value
        else:
            term, positions = itermerge.next(self)

        # Look at the next item to see if it is has positions for the current
        # term.

        try:
            t, p = itermerge.next(self)
            while t == term:
                positions = self.combine(positions, p)
                t, p = itermerge.next(self)
            self.next_value = t, p

        # Where an item could not be fetched, cause future requests to fail.

        except StopIteration:
            self.next_value = None

        return term, positions

    def close(self):
        for reader in self.readers:
            reader.close()
        self.readers = []

class Index:

    "An inverted index solution encapsulating the various components."

    def __init__(self, pathname, flush_interval=FLUSH_INTERVAL,
        open_partitions=OPEN_PARTITIONS):

        self.pathname = pathname
        self.flush_interval = flush_interval
        self.open_partitions = open_partitions
        self.reader = None
        self.writer = None

    def get_writer(self):

        "Return a writer."

        if self.writer is None:
            self._ensure_directory()
            self.writer = IndexWriter(self.pathname, self.flush_interval)
        return self.writer

    def _ensure_directory(self):
        if not exists(self.pathname):
            mkdir(self.pathname)

    def get_reader(self, refresh=0):

        "Return a reader for the index."

        if refresh and self.reader is not None:
            self.reader.close()
            self.reader = None

        if self.reader is None:
            if not exists(self.pathname):
                raise OSError, "Index path %r does not exist." % self.pathname
            self.reader = IndexReader(self.pathname)
        return self.reader

    def merge(self):

        "Merge the partitions in the index."

        reader = IndexReader(self.pathname, get_term_data_reader, self.merge_data)
        writer = get_term_writer(self.pathname, "merged")
        try:
            writer.begin(*reader.get_sizes())
            for term, data in reader:
                writer.write_term_plus_remaining(term, data)
                writer.end_record()
        finally:
            writer.close()
            reader.close()

        for partition in get_term_partitions(self.pathname):
            remove_term_files(self.pathname, partition)

        rename_term_files(self.pathname, "merged", 0)

    def merge_data(self, a, b):

        """
        Merge 'a' and 'b', modifying the data to permit concatenation.
        """

        # Modify the record to indicate a continuation of the data.

        c = a + b
        c[len(a) - 1] = 1
        return c

    def close(self):
        if self.reader is not None:
            self.reader.close()
            self.reader = None
        if self.writer is not None:
            self.writer.close()
            self.writer = None

# vim: tabstop=4 expandtab shiftwidth=4
