#!/usr/bin/env python

"""
Specific classes for storing term information.

Copyright (C) 2009, 2010, 2011 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from iixr.data import *
from iixr.files import *
from iixr.phrases import PhraseIterator
from os.path import commonprefix # to find common string prefixes

class TermWriter(FileWriter):

    "Writing term information to files."

    def begin(self, docnum_size, position_size):

        "Begin writing to the file."

        self.write_numbers((docnum_size, position_size))
        self.end_record()

        self.data_start = self.tell()
        self.docnum_size = docnum_size
        self.position_size = position_size
        self.subtractor = get_subtractor(docnum_size)
        self.last_term = ""

    def write_terms(self, terms):

        """
        Write the 'terms' to the term information file, with each term's details
        stored in a separate record.
        """

        if hasattr(terms, "items"):
            terms = terms.items()
            terms.sort()

        for term, doc_positions in terms:
            if not doc_positions:
                continue

            if hasattr(doc_positions, "items"):
                doc_positions = doc_positions.items()

            docnum, positions = doc_positions[0]

            if not positions:
                continue

            # Start the writing, if appropriate.

            if self.data_start is None:
                self.begin(sizeof(docnum), sizeof(positions[0]))

            # Write each term and document positions.

            self.write_term(term, doc_positions)
            self.end_record()

    # Methods requiring an open record.

    def write_term(self, term, doc_positions):

        """
        Write the given 'term', its document frequency (number of documents in
        which it appears), and 'doc_positions' to the term information file.
        """

        self.write_term_only(term)

        # Write the document frequency and the term positions.

        self.write_positions(doc_positions)

    def write_term_plus_remaining(self, term, data):

        "Write the given 'term' and the document position 'data'."

        self.write_term_only(term)
        self.write_remaining(data)

    def write_term_only(self, term):

        "Write only the given 'term'."

        if term <= self.last_term:
            raise ValueError, "Term %r precedes the previous term %r." % (term, self.last_term)

        # Write the prefix length and term suffix.

        common = len(commonprefix([self.last_term, term]))
        suffix = term[common:]

        self.write_number(common)
        self.write_string(suffix)

        self.last_term = term

    def write_positions(self, doc_positions):

        "Write the given 'doc_positions' to the file."

        # Make sure that the positions are sorted.

        doc_positions.sort()

        # Write the document frequency.

        self.write_number(len(doc_positions))

        last_docnum = None

        for docnum, positions in doc_positions:

            # Store the first document number as it is.

            if last_docnum is None:
                docnum_seq = docnum

            # Reject out-of-order documents.

            elif docnum < last_docnum:
                raise ValueError, "Document number %r is less than previous number %r." % (docnum, last_docnum)

            # Calculate an ongoing delta.

            else:
                docnum_seq = self.subtractor(docnum, last_docnum)

            # Write the document number and positions.

            self.write_sequence_value(docnum_seq, self.docnum_size)
            self.write_monotonic_sequence(positions, self.position_size)

            last_docnum = docnum

        # Write a terminating byte to indicate that no more document pages
        # exist.

        self.write_byte(0)

class TermReader(FileReader):

    "Reading term information from files."

    def begin(self):

        "Begin reading from the file."

        self.begin_record()
        try:
            self.docnum_size, self.position_size = self.read_numbers(2)
        except EOFError:
            self.docnum_size, self.position_size = 0, 0 # NOTE: No positions!

        self.data_start = self.tell()
        self.adder = get_adder(self.docnum_size)
        self.last_term = ""

    def get_sizes(self):
        return self.docnum_size, self.position_size

    # Methods requiring an open record.

    def read_term(self):

        "Read a term and its document positions from the term information file."

        # Read the term.

        self.read_term_only()

        # Read the document frequency and the term positions.

        positions = self.read_positions()

        return self.last_term, positions

    def read_term_plus_remaining(self):

        """
        Read a term and the unprocessed document position data.
        """

        self.read_term_only()
        return self.last_term, self.read_remaining()

    def read_term_only(self):

        "Read a term only."

        # Read the prefix length and term suffix.

        common = self.read_number()
        suffix = self.read_string()

        self.last_term = self.last_term[:common] + suffix
        return self.last_term

    def read_positions(self):

        "Read document positions from the term information file."

        doc_positions = []

        while 1:

            # Read the document frequency.

            npositions = self.read_number()

            last_docnum = None
            i = 0
            while i < npositions:

                # Read the document number.

                docnum = self.read_sequence_value(self.docnum_size)
                if last_docnum is not None:
                    docnum = self.adder(docnum, last_docnum)

                # Read the positions.

                positions = self.read_monotonic_sequence(self.position_size)
                doc_positions.append((docnum, positions))

                last_docnum = docnum
                i += 1

            # Read a terminating byte to discover whether more document pages
            # exist.

            if not self.read_byte():
                break

        return doc_positions

class TermIterator(TermReader):

    "An iterator over terms and positions read from a file."

    def __iter__(self):
        return self

    def next(self):
        try:
            self.begin_record()
            return self.read_term()
        except EOFError:
            raise StopIteration

class TermDataIterator(TermReader):

    "An iterator over terms and unprocessed document positions data."

    def __iter__(self):
        return self

    def next(self):
        try:
            self.begin_record()
            return self.read_term_plus_remaining()
        except EOFError:
            raise StopIteration

# vim: tabstop=4 expandtab shiftwidth=4
